import typer

from .config import delete_var_config


def delete_system_prompt(args):
    name = args["name"]
    if not name:
        name = typer.prompt("Name")

    delete_var_config("system_prompts", name)
    typer.secho(f"Deleted system prompt: {name}", fg=typer.colors.BRIGHT_GREEN, bold=True)



import json
import os

import typer


def show_config():
    config = load_config()
    from rich import print

    print(json.dumps(config, indent=4))


def load_config():
    config_path = os.path.expanduser("~/.mirageml.json")
    curr_config = {}
    if os.path.exists(config_path):
        with open(config_path, "r") as f:
            curr_config = json.load(f)
    curr_config = {
        "local_mode": curr_config.get("local_mode", False),
        "model": curr_config.get("model", "gpt-4"),
        "keep_files_local": curr_config.get("keep_files_local", True),
        "local": curr_config.get("local", []),
        "remote": curr_config.get("remote", []),
        "system_prompts": curr_config.get("system_prompts", []),
        "custom_models": curr_config.get("custom_models", []),
        "openai_key": curr_config.get("openai_key", ""),
    }
    save_config(curr_config)
    return curr_config


def save_config(config):
    config_path = os.path.expanduser("~/.mirageml.json")
    with open(config_path, "w") as f:
        json.dump(config, f, indent=4)


def set_config():
    config = load_config()

    choices = ["gpt-3.5-turbo", "gpt-4", "claude"] + config["custom_models"]

    valid = {
        "model": (choices, str),
        "local_mode": (("True (not recommended without GPU)", "False"), json.loads),
        "keep_files_local": (("True", "False"), json.loads),
        "openai_key": ([], str),
    }

    for key, value in config.items():
        if key not in valid:
            continue
        curvalue = value
        while True:
            options = f"[{', '.join(valid[key][0])}]" if valid[key][0] else ""
            question = f"Enter the value for '{key}' {options} (current value: {curvalue}): "
            value = input(question)
            if value == "":
                break
            if value not in [x.split()[0] for x in valid[key][0]]:
                options = f"{[x.split()[0] for x in valid[key][0]]}" if valid[key][0] else ""
                typer.secho(
                    f"Invalid value for '{key}'. Please use one of these options {options}",
                    fg=typer.colors.BRIGHT_RED,
                )
                continue
            config[key] = valid[key][1](value.lower())
            save_config(config)
            break

    typer.secho("MirageML Config updated!", fg=typer.colors.BRIGHT_GREEN)
    show_config()


def set_var_config(json_data):
    config = load_config()
    for key, value in json_data.items():
        config[key] = value
    save_config(config)


def append_var_config(key, value):
    config = load_config()
    if key not in config:
        config[key] = [value]
    else:
        config[key].append(value)
    save_config(config)


def delete_var_config(key, value):
    config = load_config()
    if key not in config:
        return
    filtered_values = [x for x in config[key] if x["name"] != value]
    config[key] = filtered_values
    save_config(config)



import os

from .config import load_config
from .utils.vectordb import create_local_qdrant_db, create_remote_qdrant_db


def fix_name(name):
    if name.startswith("http"):
        name = "_".join(name.split("/")[1:])

    name = name.replace(" ", "_")
    name = name.replace("/", "_")
    if name.startswith("_") or name.startswith("/"):
        name = name[1:]
    return name


def add_web_source(link, name=None, remote=False):
    config = load_config()
    remote = False if config["local_mode"] else True

    if not name:
        name = fix_name(link)
    else:
        name = fix_name(name)

    if remote:
        create_remote_qdrant_db(collection_name=name, link=link)
    else:
        print(f"Indexing {link}...")
        create_local_qdrant_db(collection_name=name, link=link)
    return name


def add_local_source(path=None, name=None):
    if path == "local":
        path = "."
    print("Indexing Local Files...")

    # collection_name should be absolute path
    collection_name = os.path.abspath(path) if name is None else name
    collection_name = fix_name(collection_name)

    create_local_qdrant_db(collection_name=collection_name, path=path)
    return collection_name


def add_source(name, link):
    name = fix_name(name)

    if link:
        add_web_source(link, name)
    else:
        add_local_source()



import sys

import typer


def get_sources():
    from .utils.vectordb import list_local_qdrant_db, list_remote_qdrant_db

    return list_local_qdrant_db(), list_remote_qdrant_db()


def set_sources(local_sources=None, remote_sources=None):
    from .config import set_var_config
    from .utils.vectordb import list_local_qdrant_db, list_remote_qdrant_db

    if not local_sources:
        local_sources = list_local_qdrant_db()
    if not remote_sources:
        remote_sources = list_remote_qdrant_db()

    set_var_config({"local": local_sources, "remote": remote_sources})


def list_sources():
    local_sources, remote_sources = get_sources()

    if len(local_sources) == 0 and len(remote_sources) == 0:
        typer.secho(
            f"No sources found. Please add a source using {sys.argv[0].split('/')[-1]} add source",
            fg=typer.colors.RED,
            bold=True,
        )
        return

    if len(local_sources) != 0:
        typer.secho("Local Sources:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        print("* " + "\n* ".join(local_sources))

    if len(remote_sources) != 0:
        if len(local_sources) != 0:
            print("------------------")
        typer.secho("Remote Sources:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        print("* " + "\n* ".join(remote_sources))

    set_sources(local_sources, remote_sources)



import os
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import typer
from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern


def load_gitignore_patterns(gitignore_path):
    with open(gitignore_path, "r") as f:
        lines = f.readlines()

    return [line.strip() for line in lines if line.strip() and not line.startswith("#")]


def is_ignored(path, spec):
    return spec.match_file(path)


def get_unignored_files(start_dir):
    gitignore_path = Path(start_dir) / ".gitignore"
    if not gitignore_path.exists():
        return [f for f in Path(start_dir).rglob("*") if f.is_file()]

    patterns = load_gitignore_patterns(gitignore_path)
    spec = PathSpec.from_lines(GitWildMatchPattern, patterns)

    unignored_files = []
    for root, dirs, files in os.walk(start_dir):
        # Remove dirs that are ignored to avoid traversing them
        dirs[:] = [d for d in dirs if not is_ignored(os.path.join(root, d), spec)]

        for file in files:
            filepath = os.path.join(root, file)
            if not is_ignored(filepath, spec) and not filepath.split("/")[1].startswith(".") and ".git" not in filepath:
                unignored_files.append(filepath)

    final_files = [os.path.abspath(f) for f in unignored_files]
    return final_files


def read_file(filepath):
    try:
        with open(filepath, "r", encoding="utf-8") as file:
            return file.read(), str(filepath)
    except Exception:
        pass


def crawl_files(start_dir="."):
    file_data = []

    # Walk through the directory structure
    if os.path.isfile(start_dir):
        try:
            # Read the file content
            with open(start_dir, "r", encoding="utf-8") as file:
                file_content = file.read()
                file_data.append((file_content, start_dir))
        except Exception:
            # If unable to read a file, you can print an error or continue to the next file
            typer.secho(f"Unable to read file: {start_dir}", fg=typer.colors.BRIGHT_RED, bold=True)
            return None, None
    elif not os.path.isdir(start_dir):
        typer.secho(f"Unable to read dir: {start_dir}", fg=typer.colors.BRIGHT_RED, bold=True)
        return None, None
    else:
        # Get .gitignore file content in all directories within the root directory.
        start_dir = os.path.abspath(start_dir)
        all_files = get_unignored_files(start_dir)

        with ThreadPoolExecutor() as executor:
            file_data = list(executor.map(read_file, all_files))

        file_data = [x for x in file_data if x is not None]

    data = [x[1] + ": " + x[0] for x in file_data]
    metadata = [dict({"data": x[0]}, **{"source": x[1]}) for x in file_data]
    return data, metadata



import json
import os
import uuid

import keyring
import requests
import typer
from qdrant_client import QdrantClient
from qdrant_client.http.models import Distance, PointStruct, VectorParams
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress

from ...constants import (
    SERVICE_ID,
    VECTORDB_CREATE_WEB_COLLECTION,
    VECTORDB_DELETE_ENDPOINT,
    VECTORDB_LIST_ENDPOINT,
    VECTORDB_SEARCH_ENDPOINT,
    get_headers,
)
from ..list_sources import set_sources
from .llm import _chunk_data, local_get_embedding
from .local_source import crawl_files
from .web_source import crawl_website

PACKAGE_DIR = os.path.dirname(__file__)

progress = Progress()


def get_local_qdrant_db():
    QDRANT_LOCKFILE_PATH = os.path.join(PACKAGE_DIR, ".lock")
    if os.path.exists(QDRANT_LOCKFILE_PATH):
        os.remove(QDRANT_LOCKFILE_PATH)
    return QdrantClient(path=PACKAGE_DIR)


def exists_qdrant_db(collection_name="test"):
    qdrant_client = get_local_qdrant_db()
    return collection_name in [x.name for x in qdrant_client.get_collections().collections]


def create_remote_qdrant_db(collection_name, link=None, path=None):
    user_id = keyring.get_password(SERVICE_ID, "user_id")
    console = Console()
    with Live(
        Panel("Creating Embedding....", title="[bold green]Indexer[/bold green]", border_style="green"),
        console=console,
        transient=True,
        auto_refresh=True,
        vertical_overflow="visible",
    ) as live:
        json_data = {
            "user_id": user_id,
            "collection_name": collection_name,
            "url": link,
        }
        response = requests.post(VECTORDB_CREATE_WEB_COLLECTION, json=json_data, headers=get_headers(), stream=True)
        if response.status_code == 200:
            for chunk in response.iter_lines():
                link = chunk.decode("utf-8")
                live.update(Panel(f"{link}", title="[bold green]Indexer[/bold green]", border_style="green"))

    typer.secho(f"Created Source: {collection_name}", fg=typer.colors.GREEN, bold=True)
    set_sources()
    return True


def create_local_qdrant_db(collection_name="test", link=None, path=None):
    data, metadata = [], []
    if link:
        data, metadata = crawl_website(link)
    elif path:
        data, metadata = crawl_files(path)

    qdrant_client = get_local_qdrant_db()

    final_data, final_metadata = [], []

    console = Console()
    with Live(
        Panel(
            "Creating Embeddings...",
            title="[bold green]Indexer[/bold green]",
            border_style="green",
        ),
        console=console,
        transient=True,
    ) as live:
        # For each data chunk it based on number of tokens
        final_data, final_metadata, vectors = [], [], []
        for dat, meta in zip(data, metadata):
            live.update(
                Panel(
                    f"Embedding: {meta['source']}",
                    title="[bold green]Indexer[/bold green]",
                    border_style="green",
                ),
                refresh=True,
            )
            chunk_data, chunk_meta, chunk_vec = _chunk_data(dat, meta)
            final_data.extend(chunk_data)
            final_metadata.extend(chunk_meta)
            vectors.extend(chunk_vec)

        live.update(
            Panel(
                "Creating Index...",
                title="[bold green]Indexer[/bold green]",
                border_style="green",
            ),
            refresh=True,
        )

        qdrant_client.recreate_collection(
            collection_name=collection_name,
            vectors_config=VectorParams(size=768, distance=Distance.COSINE),
        )

        for vector, f_metadata in zip(vectors, final_metadata):
            filepath = f_metadata["source"]

            live.update(
                Panel(
                    f"Indexing: {filepath}",
                    title="[bold green]Indexer[/bold green]",
                    border_style="green",
                ),
                refresh=True,
            )

            qdrant_client.upsert(
                collection_name=collection_name,
                points=[PointStruct(vector=vector, payload=f_metadata, id=uuid.uuid4().hex)],
            )

    typer.secho(f"Created Source: {collection_name}", fg=typer.colors.GREEN, bold=True)

    set_sources()
    return qdrant_client


def list_remote_qdrant_db():
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, "user_id"),
    }
    response = requests.post(VECTORDB_LIST_ENDPOINT, json=json_data, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    return response.json()


def list_local_qdrant_db():
    QDRANT_JSON_PATH = os.path.join(PACKAGE_DIR, "meta.json")
    if os.path.exists(QDRANT_JSON_PATH):
        with open(QDRANT_JSON_PATH) as json_file:
            collection = json.load(json_file)
            collection_names = list(collection["collections"].keys())
    else:
        collection_names = []
    return collection_names


def remote_qdrant_search(source_name, user_input, data=None, metadata=None):
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, "user_id"),
        "collection_name": source_name,
        "search_query": user_input,
        "data": data,
        "metadata": metadata,
    }
    response = requests.post(VECTORDB_SEARCH_ENDPOINT, json=json_data, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    set_sources()
    return response.json()


def local_qdrant_search(source_name, user_input):
    qdrant_client = get_local_qdrant_db()

    query_vector = local_get_embedding([user_input])[0]

    hits = qdrant_client.search(
        limit=5,
        collection_name=source_name,
        query_vector=query_vector,
    )
    hits = [{"score": hit.score, "payload": hit.payload} for hit in hits]
    return hits


def transient_qdrant_search(user_input, data, metadata):
    qdrant_client = QdrantClient(location=":memory:")
    collection_name = str(uuid.uuid4().hex)

    search_vector = local_get_embedding([user_input])[0]

    qdrant_client.recreate_collection(
        collection_name=collection_name, vectors_config=VectorParams(size=768, distance=Distance.COSINE)
    )

    final_data, final_metadata, vectors = [], [], []
    for dat, meta in zip(data, metadata):
        chunk_data, chunk_meta, chunk_vec = _chunk_data(dat, meta)
        final_data.extend(chunk_data)
        final_metadata.extend(chunk_meta)
        vectors.extend(chunk_vec)

    qdrant_client.upsert(
        collection_name=collection_name,
        points=[
            PointStruct(vector=vector, payload=f_metadata, id=uuid.uuid4().hex)
            for vector, f_metadata in zip(vectors, final_metadata)
        ],
    )

    limit = 20

    hits = qdrant_client.search(
        collection_name=collection_name,
        query_vector=search_vector,
        limit=limit,
    )
    hits = [{"score": hit.score, "payload": hit.payload} for hit in hits]
    return hits


def delete_remote_qdrant_db(collection_name="test"):
    json_data = {
        "user_id": keyring.get_password(SERVICE_ID, "user_id"),
        "collection_name": collection_name,
    }
    response = requests.post(VECTORDB_DELETE_ENDPOINT, json=json_data, headers=get_headers())
    response.raise_for_status()  # Raise an exception if the request failed
    set_sources()
    return response.json()


def delete_local_qdrant_db(collection_name="test"):
    qdrant_client = get_local_qdrant_db()
    qdrant_client.delete_collection(collection_name=collection_name)
    set_sources()



from prompt_toolkit import prompt
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

# Create custom key bindings
kb = KeyBindings()


# You can put your custom finish conditon function here
@kb.add(Keys.ControlJ)
def _(event):
    buffer = event.current_buffer
    buffer.insert_text("\n")


@kb.add(Keys.ControlC)
def _(event):
    buffer = event.current_buffer
    if len(buffer.text) == 0:
        event.app.exit(exception=KeyboardInterrupt, style="class:aborting")
    else:
        buffer.text = ""


@kb.add(Keys.Enter)
def _(event):
    event.current_buffer.validate_and_handle()


def multiline_input(message):
    import typer

    typer.secho(
        f"{message} (Use Control+J' for newline):",
        fg=typer.colors.BRIGHT_GREEN,
        bold=True,
    )
    user_input = prompt("> ", multiline=True, key_bindings=kb)
    return user_input


# if __name__ == '__main__':
#     user_input = multiline_input("test_file")
#     print(user_input)


def interrupted(signum, frame):
    # when receive a timeout signal,
    raise Exception("end of time")


def input_or_timeout(prompt, default="no", timeout=10):
    import signal

    # set the signal function handler
    # set the alarm after timeout seconds
    signal.signal(signal.SIGALRM, interrupted)
    signal.alarm(timeout)

    try:
        answer = input(prompt)
        signal.alarm(0)
        return answer
    except Exception:
        return default



import sys

import tiktoken
import typer
from rich.box import HORIZONTALS
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel

from .config import load_config
from .utils.custom_inputs import multiline_input
from .utils.llm import llm_call
from .utils.prompt_templates import RAG_TEMPLATE
from .utils.vectordb import (
    list_local_qdrant_db,
    list_remote_qdrant_db,
    local_qdrant_search,
    remote_qdrant_search,
    transient_qdrant_search,
)

console = Console()
config = load_config()


def search(user_input, sources, transient_sources=None, live=None):
    from concurrent.futures import ThreadPoolExecutor

    hits = []
    local = list_local_qdrant_db()
    remote = list_remote_qdrant_db()

    local_sources = [source for source in sources if source in local]
    remote_sources = [source for source in sources if source in remote]

    for source_name in local_sources:
        if live:
            live.update(
                Panel(
                    "Searching through local sources...",
                    title="[bold blue]Assistant[/bold blue]",
                    border_style="blue",
                )
            )
        # Search for matches in each source based on user input
        hits.extend(local_qdrant_search(source_name, user_input))

    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(remote_qdrant_search, source_name, user_input) for source_name in remote_sources]
        if live:
            live.update(
                Panel(
                    "Searching through remote sources...",
                    title="[bold blue]Assistant[/bold blue]",
                    border_style="blue",
                )
            )

        for future in futures:
            try:
                hits.extend(future.result())
            except Exception:
                error_msg = f"Failed to search in source: {source_name}. Try again! You may need to re-add the source with mirage add source"
                typer.secho(
                    error_msg,
                    fg=typer.colors.RED,
                    bold=True,
                )

    if transient_sources:
        if live:
            live.update(
                Panel(
                    "Searching through files and urls...",
                    title="[bold blue]Assistant[/bold blue]",
                    border_style="blue",
                )
            )
        # TODO: Optimize this
        if config["local_mode"] or config["keep_files_local"]:
            for data, metadata in transient_sources:
                hits.extend(transient_qdrant_search(user_input, data, metadata))
                if live:
                    live.update(
                        Panel(
                            "Searching through files and urls locally...",
                            title="[bold blue]Assistant[/bold blue]",
                            border_style="blue",
                        )
                    )
        else:
            source_name = "transient"
            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(remote_qdrant_search, source_name, user_input, data, metadata)
                    for data, metadata in transient_sources
                ]

                for future in futures:
                    hits.extend(future.result())

    return hits


def rank_hits(hits):
    # Rank the hits based on their relevance
    num_hits = 5
    if config["model"] == "claude":
        num_hits = 40
    elif config["model"] == "gpt-3.5-turbo":
        num_hits = 20
    elif config["model"] == "gpt-4":
        num_hits = 10

    sorted_hits = sorted(hits, key=lambda x: x["score"], reverse=True)[:num_hits]
    return sorted_hits


def create_context(sorted_hits):
    return "\n\n".join([str(x["payload"]["source"]) + ": " + x["payload"]["data"] for x in sorted_hits])


def search_and_rank(user_input, sources, transient_sources, live):
    hits = search(user_input, sources, transient_sources, live)
    sorted_hits = rank_hits(hits)
    return sorted_hits


def rag_chat(sources, transient_sources):
    try:
        all_sources = sources + [x[1][0]["source"] for x in transient_sources]
        user_input = multiline_input(f"Ask a question over these sources ({', '.join(all_sources)})")
        if user_input.lower().strip() == "exit":
            typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
            sys.exit()
    except KeyboardInterrupt:
        typer.secho("Ending chat. Goodbye!", fg=typer.colors.BRIGHT_GREEN, bold=True)
        sys.exit()

    # Live display while searching for relevant sources
    with Live(
        Panel(
            "Searching through the relevant sources...",
            title="[bold blue]Assistant[/bold blue]",
            border_style="blue",
        ),
        console=console,
        transient=True,
        auto_refresh=True,
        refresh_per_second=8,
    ) as live:
        transient_data = ""
        tsources = []
        enc = tiktoken.get_encoding("cl100k_base")
        for data, metadata in transient_sources:
            transient_data += "\n\n" + data[0]
            tsources.append(metadata[0]["source"])
        if len(enc.encode(transient_data)) < 75000:
            transient_sources = None

        sorted_hits = search_and_rank(user_input, sources, transient_sources, live)
        sources_used = list(set([hit["payload"]["source"] for hit in sorted_hits]))
        context = create_context(sorted_hits)

        if not transient_sources and len(transient_data) > 0:
            context += transient_data
            sources_used.extend(tsources)

        # Chat history that will be sent to the AI model
        chat_history = [
            {
                "role": "system",
                "content": "You are a helpful assistant. When responding to questions, provide answers concisely using the following format:\n{answer}\n\nSources:\n{sources}",
            },
            {
                "role": "user",
                "content": RAG_TEMPLATE.format(context=context, question=user_input, sources=sources_used),
            },
        ]

        # Fetch the AI's response
        ai_response = ""
        try:
            if live:
                live.update(
                    Panel(
                        "Found relevant sources! Answering question...",
                        title="[bold blue]Assistant[/bold blue]",
                        box=HORIZONTALS,
                        border_style="blue",
                    )
                )
            response = llm_call(
                chat_history,
                model=config["model"],
                stream=True,
                local=config["local_mode"],
                live=live,
            )

            if config["local_mode"]:
                for chunk in response:
                    ai_response += chunk
                    live.update(
                        Panel(
                            Markdown(ai_response),
                            title="[bold blue]Assistant[/bold blue]",
                            box=HORIZONTALS,
                            border_style="blue",
                        )
                    )
            else:
                for chunk in response.iter_content(chunk_size=512):
                    if chunk:
                        decoded_chunk = chunk.decode("utf-8")
                        ai_response += decoded_chunk
                        live.update(
                            Panel(
                                Markdown(ai_response),
                                title="[bold blue]Assistant[/bold blue]",
                                box=HORIZONTALS,
                                border_style="blue",
                            )
                        )
        except KeyboardInterrupt:
            pass
    return chat_history, ai_response



"""
This file is the entry point for the commands module.
It initializes the module and imports the necessary modules.
"""
from .add_model import add_model
from .add_source import add_source
from .add_system_prompt import add_system_prompt
from .chat import chat
from .config import set_config, show_config
from .delete_source import delete_source
from .delete_system_prompt import delete_system_prompt
from .list_models import list_models
from .list_sources import list_sources
from .list_system_prompts import list_system_prompts
from .login import login
from .profile import profile
from .tutorial import tutorial

__all__ = [
    "login",
    "profile",
    "set_config",
    "show_config",
    "chat",
    "list_system_prompts",
    "list_models",
    "list_sources",
    "add_model",
    "add_system_prompt",
    "add_source",
    "delete_source",
    "delete_system_prompt",
    "tutorial",
]



import json

import requests
import typer
from rich.console import Console
from rich.live import Live
from rich.panel import Panel

from ...constants import WEB_SCRAPE_EXTRACT_ENDPOINT, WEB_SCRAPE_LINKS_ENDPOINT, get_headers
from .custom_inputs import input_or_timeout

console = Console()


def validate_all_scraped(visited_links, urls):
    # pretty print all of the subpages that were indexed and ask the user if they want to continue
    typer.secho("Subpaths Per URL:", fg=typer.colors.GREEN, bold=True)
    print()

    # Process each link and add to tree
    for url in visited_links:
        # Find the key in urls that the visited link is under
        url_key = next((key for key in urls.keys() if key in url), None)

        if url_key:
            try:
                url = url.replace(url_key, "").split("/")[0]
                if url:
                    urls[url_key].add(url)
            except Exception:
                continue

    # Print the results
    for domain, paths in urls.items():
        typer.secho(f"{domain}:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        if len(paths) == 0:
            print("No subpaths found, you may have to manually add the url when prompted.\n")
            continue
        for path in sorted(paths):
            print(path)
        print()

    # Will timeout if user doesn't respond within 10 seconds
    user_input = input_or_timeout(
        "Do you want to index another URL? Enter a URL or leave empty (timeout: 10s) (default: no): ",
        default="no",
        timeout=10,
    )
    user_input = user_input.strip()
    if user_input and not user_input.lower().startswith("n"):
        link = user_input if user_input.lower().startswith("http") else input("Link for the source (exit to skip): ")

        while not link.lower().startswith("http") and link.lower().strip() != "exit":
            link = input("Please enter a valid link starting with http:// or type 'exit' to skip: ")

        to_visit = [link + "/" if not link.endswith("/") else link] if link.lower().strip() != "exit" else []
        if to_visit:
            urls[to_visit[0]] = set()
            start_url = to_visit[0]
    else:
        to_visit = []
        start_url = ""
    return start_url, to_visit, urls


def crawl_website(start_url):
    if not start_url.endswith("/"):
        start_url += "/"
    urls = {start_url: set()}
    visited_links = set()
    to_visit = [start_url]

    while to_visit:
        with Live(
            Panel(
                "Preparing to Scrape...",
                title="[bold green]Scraper[/bold green]",
                border_style="green",
            ),
            console=console,
            transient=True,
            auto_refresh=True,
            vertical_overflow="visible",
        ) as live:
            response = requests.post(
                WEB_SCRAPE_LINKS_ENDPOINT, json={"urls": [start_url]}, headers=get_headers(), stream=True
            )
            if response.status_code == 200:
                for chunk in response.iter_lines():
                    # process line here
                    link = chunk.decode("utf-8")
                    if live:
                        live.update(
                            Panel(
                                f"Scraping: {link}",
                                title="[bold green]Scraper[/bold green]",
                                border_style="green",
                            )
                        )
                    visited_links.add(link.strip())

        # Skip for now, assume that the scraper can find all paths
        # start_url, to_visit, urls = validate_all_scraped(visited_links, urls)
        start_url, to_visit, urls = "", [], urls

    data, metadata = [], []
    with Live(
        Panel(
            "Preparing to Clean...",
            title="[bold green]Cleaner[/bold green]",
            border_style="green",
        ),
        console=console,
        transient=True,
        auto_refresh=True,
        vertical_overflow="visible",
    ) as live:
        response = requests.post(
            WEB_SCRAPE_EXTRACT_ENDPOINT, json={"urls": list(visited_links)}, headers=get_headers(), stream=True
        )
        if response.status_code == 200:
            for chunk in response.iter_lines():
                # process line here
                decoded_chunk = chunk.decode("utf-8")
                json_data = json.loads(decoded_chunk)

                source = json_data["source"]
                data.extend([json_data["data"]])
                metadata.extend([json_data["metadata"]])

                if live:
                    live.update(
                        Panel(
                            f"Cleaning: {source}",
                            title="[bold green]Cleaner[/bold green]",
                            border_style="green",
                        )
                    )

    return data, metadata


def extract_from_url(url, live=None):
    if live:
        live.update(
            Panel(
                f"Loading: {url}",
                title="[bold blue]Assistant[/bold blue]",
                border_style="blue",
            )
        )

    response = requests.post(WEB_SCRAPE_EXTRACT_ENDPOINT, json={"urls": [url]}, headers=get_headers())
    json_data = response.json()

    source, url_data = (
        json_data["source"],
        json_data["data"],
    )
    data = [url_data]
    metadata = [{"data": url_data, "source": source}]

    return data, metadata


if __name__ == "__main__":
    import time

    start_time = time.time()
    crawl_website("https://modal.com/docs/guide/")
    print(f"Time taken: {time.time() - start_time}")



import sys

import keyring
import requests
import typer

from mirageml.constants import SERVICE_ID, SUPABASE_KEY, SUPABASE_URL


def get_models():
    access_token = keyring.get_password(SERVICE_ID, "access_token")
    user_id = keyring.get_password(SERVICE_ID, "user_id")
    headers = {
        "apikey": SUPABASE_KEY,
        "Authorization": f"Bearer {access_token}",
        "Content-Type": "application/json",
    }

    response = requests.get(
        f"{SUPABASE_URL}/rest/v1/user_finetunes?user_id=eq.{user_id}&select=model_name", headers=headers
    )

    response_data = response.json()
    model_names = []
    if len(response_data) > 0:
        model_names = [data["model_name"] for data in response_data]

    return model_names


def set_models(models=None):
    from .config import set_var_config

    if not models:
        models = get_models()

    set_var_config({"custom_models": get_models()})


def list_models():
    models = get_models()

    if len(models) == 0:
        typer.secho(
            f"No custom models found. Please add a source using {sys.argv[0].split('/')[-1]} add model",
            fg=typer.colors.RED,
            bold=True,
        )
        return

    if len(models) != 0:
        typer.secho("Custom Models:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        print("* " + "\n* ".join(models))

    set_models(models)



from supabase import create_client

SERVICE_ID = "mirageml"

PORT = 9998
REDIRECT_URI = f"http://localhost:{PORT}/callback"
USER_CHECK_ENDPOINT = "https://mirageml--miragecli-authentication-check-email.modal.run"
SUPABASE_URL = "https://kfskvbhwrwpbruczecka.supabase.co"
SUPABASE_KEY = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
    "eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6Imtmc2t2Ymh3cndwYnJ1Y"
    "3plY2thIiwicm9sZSI6ImFub24iLCJpYXQiOjE2OTcwNTI5MDYsImV4cCI6MjAxMjYyODkwNn0."
    "KZoU1QMfvh72QfKJK_HD3EQ_246g-QbggSzPbT83Su8"
)
ANALYTICS_WRITE_KEY = "WeKYF6EZAYtVxUwlx8g3sG1JKDMMv6jY"

WEB_SCRAPE_LINKS_ENDPOINT = "https://mirageml--web-scraper-recursive-links.modal.run"
WEB_SCRAPE_EXTRACT_ENDPOINT = "https://mirageml--web-scraper-extract-from-url.modal.run"

VECTORDB_EMBED_ENDPOINT = "https://mirageml--vectordb-embed-text.modal.run"
LLM_GPT_ENDPOINT = "https://mirageml--llm-cli-gpt.modal.run"

VECTORDB_SEARCH_ENDPOINT = "https://mirageml--vectordb-search-db.modal.run"
VECTORDB_LIST_ENDPOINT = "https://mirageml--vectordb-list-db.modal.run"
VECTORDB_CREATE_ENDPOINT = "https://mirageml--vectordb-create-db.modal.run"
VECTORDB_UPSERT_ENDPOINT = "https://mirageml--vectordb-upsert-db.modal.run"
VECTORDB_DELETE_ENDPOINT = "https://mirageml--vectordb-delete-db.modal.run"
VECTORDB_CREATE_WEB_COLLECTION = "https://mirageml--vectordb-create-web-collection.modal.run/"

FINETUNE_CREATE_ENDPOINT = "https://mirageml--finetune-create-model.modal.run"

supabase = create_client(SUPABASE_URL, SUPABASE_KEY)


def fetch_new_access_token():
    import keyring

    refresh_token = keyring.get_password(SERVICE_ID, "refresh_token")
    response = supabase.auth._refresh_access_token(refresh_token)
    session = response.session
    keyring.set_password(SERVICE_ID, "access_token", session.access_token)
    keyring.set_password(SERVICE_ID, "refresh_token", session.refresh_token)
    keyring.set_password(SERVICE_ID, "expires_at", str(session.expires_at))
    return session.access_token


def get_headers():
    import time

    import keyring
    import typer

    expires_at = keyring.get_password(SERVICE_ID, "expires_at")
    access_token = keyring.get_password(SERVICE_ID, "access_token")
    if expires_at and float(expires_at) < time.time():
        try:
            refresh_token = keyring.get_password(SERVICE_ID, "refresh_token")
            response = supabase.auth._refresh_access_token(refresh_token)
            session = response.session
            keyring.set_password(SERVICE_ID, "access_token", session.access_token)
            keyring.set_password(SERVICE_ID, "refresh_token", session.refresh_token)
            keyring.set_password(SERVICE_ID, "expires_at", str(session.expires_at))
            access_token = session.access_token
        except Exception as e:
            print(e)
            typer.echo("Please login again. Run `mirageml login`")
            raise typer.Exit()
    headers = {"Authorization": f"Bearer {access_token}"}
    return headers


def help_list_sources(command_prompt):
    import json
    import os
    import sys

    invoked_alias = sys.argv[0].split("/")[-1]  # Extract only the alias name

    config_path = os.path.expanduser("~/.mirageml.json")
    config = {}
    if os.path.exists(config_path):
        with open(config_path, "r") as f:
            config = json.load(f)

    local_sources = config.get("local", [])
    remote_sources = config.get("remote", [])
    all_sources = list(set(local_sources + remote_sources))
    if len(all_sources) == 0:
        final_string = (
            f"Specify sources to delete:\n\n\nEx: " f"**{invoked_alias} {command_prompt} {{source1}}**\n\n\n\n"
        )
    elif len(all_sources) == 1:
        final_string = (
            f"Specify sources to delete:\n\n\nEx: " f"**{invoked_alias} {command_prompt} {all_sources[0]}**\n\n\n\n"
        )
    else:
        command_flag = "-s " if command_prompt.split()[0] == "chat" else ""
        final_string = (
            f"Specify sources:\n\n\nEx: "
            f"**{invoked_alias} {command_prompt} {all_sources[0]} "
            f"{command_flag}{all_sources[1]}**\n\n\n\n"
        )

    if len(all_sources) != 0:
        final_string += "**Sources:**\n\n* "
        final_string += "\n\n* ".join(all_sources)
    return final_string



import sys
import time

import typer


def tutorial():
    from . import add_source, chat

    invoked_alias = sys.argv[0].split("/")[-1]

    typer.secho("\nGreat, you are now logged in.\n", fg=typer.colors.BRIGHT_GREEN)

    typer.secho(
        "Next, let's add a knowledge source for the web (ex: https://modal.com/docs)!", fg=typer.colors.BRIGHT_BLUE
    )
    while True:
        link = input("Link for the source: ")
        if not link.startswith("http"):
            typer.secho("Please enter a valid link starting with https://", fg=typer.colors.RED, bold=True)
            continue
        break
    from urllib.parse import urlparse

    parsed_url = urlparse(link)
    source_name = parsed_url.netloc.split(".")[0]
    if source_name in ["docs", "www", "en", "platform", "blog"]:
        source_name = parsed_url.netloc.split(".")[1]
    source_name = input(f"source_name for the source [default: {source_name}]: ") or source_name
    add_source(source_name, link)

    time.sleep(2)
    typer.secho("\nYou have added your first source!", fg=typer.colors.BRIGHT_GREEN)

    # typer.secho("Now, let's add a plugin for additional capabilities:", fg=typer.colors.BRIGHT_BLUE)
    # add_plugin()
    # typer.secho("\nPlugin added.\n", fg=typer.colors.BRIGHT_GREEN)

    time.sleep(2)
    typer.secho(
        f"Use `{invoked_alias} chat -s modal` when your source is ready and `{invoked_alias} list sources` to list your sources\n",
        fg=typer.colors.BRIGHT_GREEN,
    )

    time.sleep(2)
    typer.secho(
        f"Use `{invoked_alias} chat -u <link_to_webpage>` to open a chat over a specific url",
        fg=typer.colors.BRIGHT_BLUE,
    )

    time.sleep(2)
    typer.secho(
        f"Use `{invoked_alias} chat -f <file_or_directory>` to open a chat over a file or directory",
        fg=typer.colors.BRIGHT_BLUE,
    )

    time.sleep(2)
    typer.secho("Let's chat with a file or directory!", fg=typer.colors.BRIGHT_BLUE)

    from prompt_toolkit import prompt
    from prompt_toolkit.completion import PathCompleter

    file_path = prompt("Enter the filepath to a file or directory: ", completer=PathCompleter())

    chat([file_path])

    time.sleep(2)
    typer.secho("\nAnd that concludes the tutorial! You can now use MirageML for:", fg=typer.colors.BRIGHT_GREEN)

    time.sleep(1)
    typer.echo("- Logging in and managing your account")

    time.sleep(1)
    typer.echo("- Adding and managing knowledge sources")

    time.sleep(1)
    typer.echo("- Finetuning a custom model")

    time.sleep(1)
    typer.echo("- Adding plugins")

    time.sleep(1)
    typer.echo("- Chatting with your context-aware assistant")

    typer.secho(f"\nType `{invoked_alias} --help` for more info on all the commands.\n", fg=typer.colors.YELLOW)



import sys
import time
from typing import List

import typer

app = typer.Typer(
    no_args_is_help=True,
    add_completion=False,
    rich_markup_mode="markdown",
    help="""
    MirageML CLI

    **You can use 'mirage' or 'mml' to call the package.**



    See the website at https://mirageml.com/ for documentation and more information
    about running code on MirageML.
    """,
    # pretty_exceptions_short=False, pretty_exceptions_enable=False # Needed for debugging
)

config_app = typer.Typer(name="config", help="Manage the config", no_args_is_help=True)
add_app = typer.Typer(name="add", help="Add a new resource", no_args_is_help=True)
list_app = typer.Typer(name="list", help="List resources", no_args_is_help=True)
sync_app = typer.Typer(name="sync", help="Sync resources", no_args_is_help=True)
delete_app = typer.Typer(name="delete", help="Delete resources", no_args_is_help=True)
remove_app = typer.Typer(name="remove", help="Remove resources", no_args_is_help=True)


app.add_typer(config_app, rich_help_panel="Utils and Configs")
app.add_typer(add_app, rich_help_panel="Manage Resources")
app.add_typer(list_app, rich_help_panel="Manage Resources")
app.add_typer(delete_app, rich_help_panel="Manage Resources")
app.add_typer(sync_app, rich_help_panel="Manage Resources", hidden=True)
app.add_typer(remove_app, rich_help_panel="Manage Resources", hidden=True)


@app.callback()
def main(ctx: typer.Context):
    import keyring
    import segment.analytics as analytics

    from .constants import ANALYTICS_WRITE_KEY, SERVICE_ID, fetch_new_access_token

    analytics.write_key = ANALYTICS_WRITE_KEY

    user_id = keyring.get_password(SERVICE_ID, "user_id")
    expires_at = keyring.get_password(SERVICE_ID, "expires_at")
    if not user_id and (ctx.invoked_subcommand != "login"):
        typer.echo("Please login first. Run `mirageml login`")
        raise typer.Exit()
    elif expires_at and float(expires_at) < time.time() and (ctx.invoked_subcommand != "login"):
        try:
            fetch_new_access_token()
            analytics.identify(user_id)
        except Exception:
            typer.echo("Please login again. Run `mirageml login`")
            raise typer.Exit()

    if user_id:
        full_command = " ".join(sys.argv[1:])
        analytics.track(user_id, "command", {"command": full_command})


@app.command(name="help", hidden=True)
def custom_help():
    # Your custom help message
    import os

    os.system("mml --help")


@app.command(name="login")
def login_command():
    """Login to Mirage ML"""
    from .commands import login

    login()


@app.command(name="tutorial")
def tutorial():
    """Walk through the basics of using mirageml"""
    from .commands import login, tutorial

    typer.secho("Welcome to the MirageML tutorial!", fg=typer.colors.BRIGHT_GREEN, bold=True)
    typer.secho("Here, you will learn how to use the main features of MirageML.", fg=typer.colors.BRIGHT_GREEN)
    # Add more instructions here...
    typer.secho("First we will sign you up:", fg=typer.colors.BRIGHT_GREEN)
    login()
    tutorial()


@app.command(name="profile", hidden=True)
def profile_command():
    """View your profile"""
    from .commands import profile

    profile()


def generate_chat_help_text():
    from .constants import help_list_sources

    return help_list_sources("chat -s")


@app.command(name="chat")
def chat_command(
    filepaths: List[str] = typer.Option(
        None,
        "--files",
        "-f",
        help="Path to files/directories to use as context. \n\n\n**"
        + sys.argv[0].split("/")[-1]
        + " chat -f {filepath} -f {directory}**",
    ),
    urls: List[str] = typer.Option(
        None,
        "--urls",
        "-u",
        help="URLs to use as context. \n\n\n**" + sys.argv[0].split("/")[-1] + " chat -u {url1} -u {url2}**",
    ),
    sources: List[str] = typer.Option(None, "--sources", "-s", help=generate_chat_help_text()),
    sp: str = typer.Option(None, "--system-prompt", "-sp", help="Name of the system prompt to use"),
):
    """Chat with MirageML"""
    for url in urls:
        if not url.startswith("http"):
            typer.echo("Every url must start with http://")
            raise typer.Exit()

    if len(urls + filepaths) > 0:
        typer.secho(
            "Indexing sources...",
            fg=typer.colors.BRIGHT_GREEN,
            bold=True,
        )
    else:
        typer.secho(
            "Starting chat. Type 'exit' to end the chat.",
            fg=typer.colors.BRIGHT_GREEN,
            bold=True,
        )

    from .commands import chat

    chat(files=filepaths, urls=urls, sources=sources, sp=sp)


@config_app.command(name="show")
def show_config_command():
    """Show the current config"""
    from .commands import show_config

    show_config()


@config_app.command(name="get", hidden=True)
def get_config_command():
    """Show the current config"""
    from .commands import show_config

    show_config()


@config_app.command(name="set")
def set_config_command():
    """
    Set the config file for MirageML
    """
    from .commands import set_config

    set_config()


# List Commands
@list_app.command(name="sp", hidden=True)
def list_system_prompts():
    from .commands import list_system_prompts

    list_system_prompts()


@list_app.command(name="sources")
def list_sources_command():
    """List created sources"""
    from .commands import list_sources

    list_sources()


@list_app.command(name="models")
def list_models_command():
    """List created models"""
    from .commands import list_models

    list_models()


# Add System Prompt
@add_app.command(name="sp", hidden=True)
def add_system_prompt_command(name: str = typer.Argument(default="", help="Name of the system prompt")):
    """Add a new system prompt"""
    from .commands import add_system_prompt

    add_system_prompt({"name": name})


@add_app.command(name="source")
def add_source_command(link: str = typer.Argument(default="", help="Link to the source")):
    """Add a new source"""
    from .commands import add_source

    if not link.startswith("http"):
        if link:
            typer.secho("Please enter a valid link starting with https://", fg=typer.colors.RED, bold=True)
        while True:
            link = input("Link for the source: ")
            if not link.startswith("http"):
                typer.secho("Please enter a valid link starting with https://", fg=typer.colors.RED, bold=True)
                continue
            break
    from urllib.parse import urlparse

    parsed_url = urlparse(link)
    name = parsed_url.netloc.split(".")[0]
    if name in ["docs", "www", "en", "platform", "blog"]:
        name = parsed_url.netloc.split(".")[1]
    name = input(f"Name for the source [default: {name}]: ") or name
    add_source(name, link)


@add_app.command(name="sources", hidden=True)
def add_sources_command(link: str = typer.Argument(default="")):
    typer.secho("Did you mean 'add source'?", fg=typer.colors.YELLOW)
    return


@add_app.command(name="model", no_args_is_help=True)
def add_model_command(
    links: List[str] = typer.Argument(
        ...,
        help="Links to finetune on separated by spaces\n\n\nEx: mirage add model https://modal.com https://apple.com",
    )
):
    """Add a new model"""
    from .commands import add_model

    for link in links:
        if not link.startswith("http"):
            typer.secho("Please ensure links start with https://", fg=typer.colors.RED, bold=True)
            raise typer.Exit()

    from urllib.parse import urlparse

    parsed_url = urlparse(links[0])
    name = parsed_url.netloc.split(".")[0]
    if name in ["docs", "www", "en", "platform", "blog"]:
        name = parsed_url.netloc.split(".")[1]
    name = input(f"Name for the model [default: {name}]: ") or name
    add_model(name, links)


@add_app.command(name="models", hidden=True)
def add_models_command(links: List[str] = typer.Argument(...)):
    typer.secho("Did you mean 'add model'?", fg=typer.colors.YELLOW)
    return


# Delete Commands
@delete_app.command(name="sp", hidden=True)
def delete_system_prompt(name: str = typer.Argument(default="", help="Name of the system prompt")):
    from .commands import delete_system_prompt

    delete_system_prompt({"name": name})


def generate_delete_help_text():
    from .constants import help_list_sources

    return help_list_sources("delete source")


@remove_app.command(name="source", no_args_is_help=True, hidden=True)
@delete_app.command(name="source", no_args_is_help=True)
def delete_source_command(names: List[str] = typer.Argument(help=generate_delete_help_text())):
    """Delete sources"""
    from .commands import delete_source

    delete_source(names)


if __name__ == "__main__":
    app()



import json

import typer

from .config import load_config


def list_system_prompts():
    config = load_config()
    if len(config["system_prompts"]) == 0:
        typer.secho("No System Prompts. Add one with 'mirageml add sp'", fg=typer.colors.BRIGHT_RED, bold=True)
    else:
        typer.secho("System Prompts:", fg=typer.colors.BRIGHT_GREEN, bold=True)
        print(json.dumps(config["system_prompts"], indent=4))
